from dataclasses import dataclass


@dataclass(frozen=True)
class APIKeys:
    """APIキーのオブジェクト

    Attributes
    ----------
    ro_key: str
        Read Only Key
    rw_key: str
        Read-Write Key
    """
    ro_key: str
    rw_key: str



from typing import TYPE_CHECKING

import httpx

from wikidot.common.exceptions import SessionCreateException

if TYPE_CHECKING:
    from wikidot.module.client import Client


class HTTPAuthentication:
    @staticmethod
    def login(
            client: 'Client',
            username: str,
            password: str,
    ):
        """ユーザー名とパスワードでログインする

        Parameters
        ----------
        client: Client
            クライアント
        username: str
            ユーザー名
        password: str
            パスワード
        """
        # ログインリクエスト実行
        response = httpx.post(
            url='https://www.wikidot.com/default--flow/login__LoginPopupScreen',
            data={
                'login': username,
                'password': password,
                'action': 'Login2Action',
                'event': 'login'
            },
            headers=client.amc_client.header.get_header(),
            timeout=20
        )

        # Check status code
        if response.status_code != httpx.codes.OK:
            raise SessionCreateException(
                'Login attempt is failed due to HTTP status code: ' + str(response.status_code))

        # Check body
        if 'The login and password do not match' in response.text:
            raise SessionCreateException('Login attempt is failed due to invalid username or password')

        # Check cookies
        if 'WIKIDOT_SESSION_ID' not in response.cookies:
            raise SessionCreateException('Login attempt is failed due to invalid cookies')

        # Set cookies
        client.amc_client.header.set_cookie('WIKIDOT_SESSION_ID', response.cookies['WIKIDOT_SESSION_ID'])

    @staticmethod
    def logout(client: 'Client'):
        try:
            client.amc_client.request([{
                'action': 'Login2Action',
                'event': 'logout',
                'moduleName': 'Empty'
            }])
        except Exception:
            pass

        client.amc_client.header.delete_cookie('WIKIDOT_SESSION_ID')



from typing import TYPE_CHECKING

import bs4

from wikidot.module import user

if TYPE_CHECKING:
    from wikidot.module.client import Client


def user_parse(client: 'Client', elem: bs4.Tag) -> user.AbstractUser:
    """ printuser要素をパースし、ユーザーオブジェクトを返す

    Parameters
    ----------
    elem: bs4.Tag
        パース対象の要素（printuserクラスがついた要素）
    client: Client
        クライアント

    Returns
    -------
    user.AbstractUser
        パースされて得られたユーザーオブジェクト
        User | DeletedUser | AnonymousUser | GuestUser | WikidotUser のいずれか
    """

    if "class" in elem.attrs and "deleted" in elem["class"]:
        return user.DeletedUser(
            client=client,
            id=int(elem["data-id"])
        )

    elif "class" in elem.attrs and "anonymous" in elem["class"]:
        ip = elem.find("span", class_="ip").get_text().replace("(", "").replace(")", "").strip()
        return user.AnonymousUser(
            client=client,
            ip=ip
        )

    # TODO: [[user ukwhatn]]構文をパースできなくなる（aが1つしかない）ので、一度無効化 -> GuestUserの例を探して実装を戻す
    # elif len(elem.find_all("a", recursive=False)) == 1:
    #     return user.GuestUser(
    #         name=elem.get_text().strip()
    #     )

    elif elem.get_text() == "Wikidot":
        return user.WikidotUser(client=client)

    else:
        _user = elem.find_all("a")[-1]
        user_name = _user.get_text()
        user_unix = str(_user["href"]).replace("http://www.wikidot.com/user:info/", "")
        user_id = int(
            str(_user["onclick"]).replace("WIKIDOT.page.listeners.userInfo(", "").replace("); return false;", "")
        )

        return user.User(
            client=client,
            id=user_id,
            name=user_name,
            unix_name=user_unix,
            avatar_url=f"http://www.wikidot.com/avatar.php?userid={user_id}"
        )



import logging


# Logger設定
def setup_logger(name: str = 'wikidot', level=logging.INFO):
    _logger = logging.getLogger(name)
    _logger.setLevel(level)

    # ログフォーマット
    formatter = logging.Formatter('%(asctime)s [%(name)s/%(levelname)s] %(message)s')
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    _logger.addHandler(stream_handler)

    return _logger


logger = setup_logger()



# ---
# 基底クラス
# ---

class WikidotException(Exception):
    """独自例外の基底クラス"""

    def __init__(self, message):
        super().__init__(message)


# ---
# ワイルドカード
# ---

class UnexpectedException(WikidotException):
    """予期せぬ例外が発生したときの例外"""

    def __init__(self, message):
        super().__init__(message)


# ---
# セッション関連
# ---

class SessionCreateException(WikidotException):
    """セッションの作成に失敗したときの例外"""

    def __init__(self, message):
        super().__init__(message)


class LoginRequiredException(WikidotException):
    """ログインが必要なメソッドをときの例外"""

    def __init__(self, message):
        super().__init__(message)


# ---
# AMC関連
# ---
class AjaxModuleConnectorException(WikidotException):
    """ajax-module-connector.phpへのリクエストに失敗したときの例外"""

    def __init__(self, message):
        super().__init__(message)


class AMCHttpStatusCodeException(AjaxModuleConnectorException):
    """AMCから返却されたHTTPステータスが200以外だったときの例外"""

    def __init__(self, message, status_code: int):
        super().__init__(message)
        self.status_code = status_code


class WikidotStatusCodeException(AjaxModuleConnectorException):
    """AMCから返却されたデータ内のステータスがokではなかったときの例外

    HTTPステータスが200以外の場合はAMCHttpStatusCodeExceptionを投げる
    """

    def __init__(self, message, status_code: str):
        super().__init__(message)
        self.status_code = status_code


class ResponseDataException(AjaxModuleConnectorException):
    """AMCから返却されたデータが不正だったときの例外"""

    def __init__(self, message):
        super().__init__(message)


# ---
# ターゲットエラー関連
# ---

class NotFoundException(WikidotException):
    """サイトやページ・ユーザが見つからなかったときの例外"""

    def __init__(self, message):
        super().__init__(message)


class TargetErrorException(WikidotException):
    """メソッドの対象としたオブジェクトに操作が適用できないときの例外"""

    def __init__(self, message):
        super().__init__(message)


class ForbiddenException(WikidotException):
    """権限がないときの例外"""

    def __init__(self, message):
        super().__init__(message)



from datetime import datetime

import bs4


def odate_parse(odate_element: bs4.Tag) -> datetime:
    """odate要素を解析し、datetimeオブジェクトを返す

    Parameters
    ----------
    odate_element: bs4.Tag
        odate要素

    Returns
    -------
    datetime
        odate要素が表す日時

    Raises
    ------
    ValueError
        odate要素が有効なunix timeを含んでいない場合

    """
    _odate_classes = odate_element["class"]
    for _odate_class in _odate_classes:
        if "time_" in str(_odate_class):
            unix_time = int(str(_odate_class).replace("time_", ""))
            return datetime.fromtimestamp(unix_time)

    raise ValueError("odate element does not contain a valid unix time")



from dataclasses import dataclass
from typing import TYPE_CHECKING

from bs4 import BeautifulSoup

from wikidot.common import exceptions
from wikidot.common.decorators import login_required
from wikidot.util.parser import user as user_parser

if TYPE_CHECKING:
    from wikidot.module.user import AbstractUser
    from wikidot.module.site import Site


@dataclass
class SiteApplication:
    site: 'Site'
    user: 'AbstractUser'
    text: str

    def __str__(self):
        return f'SiteApplication(user={self.user}, site={self.site}, text={self.text})'

    @staticmethod
    @login_required
    def acquire_all(site: 'Site') -> list['SiteApplication']:
        """サイトへの未処理の申請を取得する

        Parameters
        ----------
        site: Site
            サイト

        Returns
        -------
        list[SiteApplication]
            申請のリスト
        """
        response = site.amc_request([{
            'moduleName': 'managesite/ManageSiteMembersApplicationsModule'
        }])[0]

        body = response.json()['body']

        if 'WIKIDOT.page.listeners.loginClick(event)' in body:
            raise exceptions.ForbiddenException('You are not allowed to access this page')

        html = BeautifulSoup(response.json()['body'], 'lxml')

        applications = []

        user_elements = html.select('h3 span.printuser')
        text_wrapper_elements = html.select('table')

        if len(user_elements) != len(text_wrapper_elements):
            raise exceptions.UnexpectedException('Length of user_elements and text_wrapper_elements are different')

        for i in range(len(user_elements)):
            user_element = user_elements[i]
            text_wrapper_element = text_wrapper_elements[i]

            user = user_parser(site.client, user_element)
            text = text_wrapper_element.select('td')[1].text.strip()

            applications.append(SiteApplication(site, user, text))

        return applications

    @login_required
    def _process(self, action: str):
        """申請を処理する

        Parameters
        ----------
        action: str
            処理の種類
        """
        if action not in ['accept', 'decline']:
            raise ValueError(f'Invalid action: {action}')

        try:
            self.site.amc_request([{
                'action': 'ManageSiteMembershipAction',
                'event': f'acceptApplication',
                'user_id': self.user.id,
                'text': f'your application has been {action}ed',
                'type': action,
                'moduleName': 'Empty'
            }])
        except exceptions.WikidotStatusCodeException as e:
            if e.status_code == 'no_application':
                raise exceptions.NotFoundException(f'Application not found: {self.user}') from e
            else:
                raise e

    def accept(self):
        """申請を承認する"""
        self._process('accept')

    def decline(self):
        """申請を拒否する"""
        self._process('decline')



from functools import wraps

from .exceptions import LoginRequiredException


def login_required(func):
    """ログインが必要な関数につけるデコレータ

    関数のパラメータに存在するclientを取得し、ログインしていない場合は例外を送出する
    clientはパラメータ、もしくはself.clientで取得できる
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        client = None
        if 'client' in kwargs:
            client = kwargs['client']
        else:
            from wikidot.module.client import Client
            for arg in args:
                if isinstance(arg, Client):
                    client = arg
                    break

            # selfに存在するか？
            if client is None:
                if hasattr(args[0], 'client'):
                    client = args[0].client

        if client is None:
            raise ValueError('Client is not found')

        if not client.is_logged_in:
            raise LoginRequiredException('Login is required to execute this function')

        return func(*args, **kwargs)

    return wrapper



from dataclasses import dataclass
from typing import TYPE_CHECKING

from bs4 import BeautifulSoup

from wikidot.common.exceptions import NotFoundException
from wikidot.util.requestutil import RequestUtil
from wikidot.util.stringutil import StringUtil

if TYPE_CHECKING:
    from wikidot.module.client import Client


class UserCollection(list):
    """ユーザーオブジェクトのリスト"""

    @staticmethod
    def from_names(
            client: 'Client',
            names: list[str],
            raise_when_not_found: bool = False
    ) -> 'UserCollection':
        """ユーザー名のリストからユーザーオブジェクトのリストを取得する

        Parameters
        ----------
        client: Client
            クライアント
        names: list[str]
            ユーザー名のリスト
        raise_when_not_found: bool
            ユーザーが見つからない場合に例外を送出するかどうか (True: 送出する, False: 送出しない)
            デフォルトでは送出しない

        Returns
        -------
        UserCollection
            ユーザーオブジェクトのリスト
        """
        responses = RequestUtil.request(
            client,
            'GET',
            [f'https://www.wikidot.com/user:info/{StringUtil.to_unix(name)}' for name in names]
        )

        users = []

        for response in responses:
            if isinstance(response, Exception):
                raise response

            html = BeautifulSoup(response.text, 'lxml')

            # 存在チェック
            if html.select_one('div.error-block'):
                if raise_when_not_found:
                    raise NotFoundException(f'User not found: {response.url}')
                else:
                    users.append(None)
                    continue

            # id取得
            user_id = int(html.select_one('a.btn.btn-default.btn-xs')['href'].split('/')[-1])

            # name取得
            name = html.select_one('h1.profile-title').get_text(strip=True)

            # avatar_url取得
            avatar_url = f'https://www.wikidot.com/avatar.php?userid={user_id}'

            users.append(User(
                client=client,
                id=user_id,
                name=name,
                unix_name=StringUtil.to_unix(name),
                avatar_url=avatar_url
            ))

        return UserCollection(users)


@dataclass
class AbstractUser:
    """ユーザーオブジェクトの抽象クラス

    Attributes
    ----------
    client: Client
        クライアントクラスのインスタンス
    id: int | None
        ユーザーID
    name: str | None
        ユーザー名
    unix_name: str | None
        ユーザーのUNIX名
    avatar_url: str | None
        ユーザーアバターのURL
    ip: str | None
        ユーザーのIPアドレス
    """
    client: 'Client'
    id: int | None = None
    name: str | None = None
    unix_name: str | None = None
    avatar_url: str | None = None
    ip: str | None = None

    def __str__(self):
        return f'{self.__class__.__name__}(id={self.id}, name={self.name}, unix_name={self.unix_name})'


@dataclass
class User(AbstractUser):
    """一般のユーザーオブジェクト

    Attributes
    ----------
    client: Client
        クライアントクラスのインスタンス
    id: int | None
        ユーザーID
    name: str | None
        ユーザー名
    unix_name: str | None
        ユーザーのUNIX名
    avatar_url: str | None
        ユーザーアバターのURL
    ip: None
        ユーザーのIPアドレス（取得できないためNone）
    """
    # client: 'Client'
    # id: int | None
    # name: str | None
    # unix_name: str | None
    # avatar_url: str | None
    ip: None = None

    @staticmethod
    def from_name(
            client: 'Client',
            name: str,
            raise_when_not_found: bool = False
    ) -> 'User':
        """ユーザー名からユーザーオブジェクトを取得する

        Parameters
        ----------
        client: Client
            クライアント
        name: str
            ユーザー名
        raise_when_not_found: bool
            ユーザーが見つからない場合に例外を送出するかどうか (True: 送出する, False: 送出しない)
            デフォルトでは送出しない

        Returns
        -------
        User
            ユーザーオブジェクト
        """
        return UserCollection.from_names(client, [name], raise_when_not_found)[0]


@dataclass
class DeletedUser(AbstractUser):
    """削除されたユーザーオブジェクト

    Attributes
    ----------
    client: Client
        クライアントクラスのインスタンス
    id: int | None
        ユーザーID
    name: str
        ユーザー名（削除されたため"account deleted"）
    unix_name: str
        ユーザーのUNIX名（削除されたため"account_deleted"）
    avatar_url: None
        ユーザーアバターのURL（削除されたためNone）
    ip: None
        ユーザーのIPアドレス（取得できないためNone）
    """
    # client: 'Client'
    # id: int | None
    name: str = "account deleted"
    unix_name: str = "account_deleted"
    avatar_url: None = None
    ip: None = None


@dataclass
class AnonymousUser(AbstractUser):
    """匿名ユーザーオブジェクト

    Attributes
    ----------
    client: Client
        クライアントクラスのインスタンス
    id: None
        ユーザーID（匿名ユーザーのためNone）
    name: str
        ユーザー名（匿名ユーザーのため"Anonymous"）
    unix_name: str
        ユーザーのUNIX名（匿名ユーザーのため"anonymous"）
    avatar_url: None
        ユーザーアバターのURL（匿名ユーザーのためNone）
    ip: str
        ユーザーのIPアドレス
    """
    # client: 'Client'
    id: None = None
    name: str = "Anonymous"
    unix_name: str = "anonymous"
    avatar_url: None = None
    # ip: None = None


@dataclass
class GuestUser(AbstractUser):
    """ゲストユーザーオブジェクト

    Attributes
    ----------
    client: Client
        クライアントクラスのインスタンス
    id: None
        ユーザーID（ゲストユーザーのためNone）
    name: str
        ユーザー名
    unix_name: None
        ユーザーのUNIX名（ゲストユーザーのためNone）
    avatar_url: None
        ユーザーアバターのURL（ゲストユーザーのためNone）
    ip: None
        ユーザーのIPアドレス（取得できないためNone）
    """
    # client: 'Client'
    id: None = None
    # name: str | None
    unix_name: None = None
    avatar_url: None = None
    ip: None = None


@dataclass
class WikidotUser(AbstractUser):
    """Wikidotシステムユーザーオブジェクト

    Attributes
    ----------
    client: Client
        クライアントクラスのインスタンス
    id: None
        ユーザーID（WikidotシステムユーザーのためNone）
    name: str
        ユーザー名（Wikidotシステムユーザーのため"Wikidot"）
    unix_name: str
        ユーザーのUNIX名（Wikidotシステムユーザーのため"wikidot"）
    avatar_url: None
        ユーザーアバターのURL（WikidotシステムユーザーのためNone）
    ip: None
        ユーザーのIPアドレス（取得できないためNone）
    """
    # client: 'Client'
    id: None = None
    name: str = "Wikidot"
    unix_name: str = "wikidot"
    avatar_url: None = None
    ip: None = None



import re

from .table import char_table


class StringUtil:
    @staticmethod
    def to_unix(target_str: str) -> str:
        """Unix形式に文字列を変換する

        Parameters
        ----------
        target_str: str
            変換対象の文字列

        Returns
        -------
        str
            変換された文字列
        """
        # MEMO: legacy wikidotの実装に合わせている

        # 特殊文字の変換辞書の作成
        table = str.maketrans(char_table.special_char_map)
        # 変換実施
        target_str = target_str.translate(table)

        # lowercaseへの変換
        target_str = target_str.lower()

        # ascii以外の文字を削除
        target_str = re.sub(r'[^a-z0-9\-:_]', '-', target_str)
        target_str = re.sub(r'^_', ':_', target_str)
        target_str = re.sub(r'(?<!:)_', '-', target_str)
        target_str = re.sub(r'^-*', '', target_str)
        target_str = re.sub(r'-*$', '', target_str)
        target_str = re.sub(r'-{2,}', '-', target_str)
        target_str = re.sub(r':{2,}', ':', target_str)
        target_str = target_str.replace(':-', ':')
        target_str = target_str.replace('-:', ':')
        target_str = target_str.replace('_-', '_')
        target_str = target_str.replace('-_', '_')

        # 先頭と末尾の:を削除
        target_str = re.sub(r'^:', '', target_str)
        target_str = re.sub(r':$', '', target_str)

        return target_str



import re
from dataclasses import dataclass
from typing import TYPE_CHECKING

import httpx

from wikidot.common import exceptions
from wikidot.common.decorators import login_required
from wikidot.module.site_application import SiteApplication

if TYPE_CHECKING:
    from wikidot.module.client import Client
    from wikidot.module.user import User


@dataclass
class Site:
    """サイトオブジェクト

    Attributes
    ----------
    id: int
        サイトID
    title: str
        サイトのタイトル
    unix_name: str
        サイトのUNIX名
    domain: str
        サイトのドメイン
    ssl_supported: bool
        SSL対応しているかどうか

    Raises
    ------
    NotFoundException
        サイトが存在しない場合
    UnexpectedException
        予期しないエラーが発生した場合
    """
    client: 'Client'

    id: int
    title: str
    unix_name: str
    domain: str
    ssl_supported: bool

    def __str__(self):
        return f'Site(id={self.id}, title={self.title}, unix_name={self.unix_name})'

    @staticmethod
    def from_unix_name(
            client: 'Client',
            unix_name: str
    ) -> 'Site':
        """UNIX名からサイトオブジェクトを取得する

        Parameters
        ----------
        client: Client
            クライアント
        unix_name: str
            サイトのUNIX名

        Returns
        -------
        Site
            サイトオブジェクト
        """
        # サイト情報を取得
        # リダイレクトには従う
        response = httpx.get(
            f'http://{unix_name}.wikidot.com',
            follow_redirects=True,
            timeout=client.amc_client.config.request_timeout
        )

        # サイトが存在しない場合
        if response.status_code == httpx.codes.NOT_FOUND:
            raise exceptions.NotFoundException(f'Site is not found: {unix_name}.wikidot.com')

        # サイトが存在する場合
        source = response.text

        # id : WIKIREQUEST.info.siteId = xxxx;
        id_match = re.search(r'WIKIREQUEST\.info\.siteId = (\d+);', source)
        if id_match is None:
            raise exceptions.UnexpectedException(f'Cannot find site id: {unix_name}.wikidot.com')
        site_id = int(id_match.group(1))

        # title : titleタグ
        title_match = re.search(r'<title>(.*?)</title>', source)
        if title_match is None:
            raise exceptions.UnexpectedException(f'Cannot find site title: {unix_name}.wikidot.com')
        title = title_match.group(1)

        # unix_name : WIKIREQUEST.info.siteUnixName = "xxxx";
        unix_name_match = re.search(r'WIKIREQUEST\.info\.siteUnixName = "(.*?)";', source)
        if unix_name_match is None:
            raise exceptions.UnexpectedException(f'Cannot find site unix_name: {unix_name}.wikidot.com')
        unix_name = unix_name_match.group(1)

        # domain :WIKIREQUEST.info.domain = "xxxx";
        domain_match = re.search(r'WIKIREQUEST\.info\.domain = "(.*?)";', source)
        if domain_match is None:
            raise exceptions.UnexpectedException(f'Cannot find site domain: {unix_name}.wikidot.com')
        domain = domain_match.group(1)

        # SSL対応チェック
        ssl_supported = str(response.url).startswith('https')

        return Site(
            client=client,
            id=site_id,
            title=title,
            unix_name=unix_name,
            domain=domain,
            ssl_supported=ssl_supported
        )

    def amc_request(self, bodies: list[dict], return_exceptions: bool = False):
        """このサイトに対してAMCリクエストを実行する"""
        return self.client.amc_client.request(bodies, return_exceptions, self.unix_name, self.ssl_supported)

    def get_applications(self):
        """サイトへの未処理の参加申請を取得する"""
        return SiteApplication.acquire_all(self)

    @login_required
    def invite_user(self, user: 'User', text: str):
        """ユーザーをサイトに招待する
        """
        try:
            self.amc_request([{
                'action': 'ManageSiteMembershipAction',
                'event': 'inviteMember',
                'user_id': user.id,
                'text': text,
                'moduleName': 'Empty'
            }])
        except exceptions.WikidotStatusCodeException as e:
            if e.status_code == 'already_invited':
                raise exceptions.TargetErrorException(
                    f'User is already invited to {self.unix_name}: {user.name}') from e
            elif e.status_code == 'already_member':
                raise exceptions.TargetErrorException(
                    f'User is already a member of {self.unix_name}: {user.name}') from e
            else:
                raise e



import asyncio
from typing import TYPE_CHECKING

import httpx

if TYPE_CHECKING:
    from wikidot.module.client import Client


class RequestUtil:
    @staticmethod
    def request(
            client: 'Client',
            method: str,
            urls: list[str],
            return_exceptions: bool = False
    ) -> list[httpx.Response | Exception]:
        """GETリクエストを送信する

        Parameters
        ----------
        client: Client
            クライアント
        method: str
            リクエストメソッド
        urls: list[str]
            URLのリスト
        return_exceptions: bool
            例外を返すかどうか (True: 返す, False: 例外を送出)
            デフォルトでは例外を送出

        Returns
        -------
        list[httpx.Response | Exception]
            レスポンスのリスト
        """
        config = client.amc_client.config
        semaphore = asyncio.Semaphore(config.semaphore_limit)

        async def _get(
                url: str
        ) -> httpx.Response:
            async with semaphore:
                async with httpx.AsyncClient() as _client:
                    return await _client.get(url)

        async def _post(
                url: str
        ) -> httpx.Response:
            async with semaphore:
                async with httpx.AsyncClient() as _client:
                    return await _client.post(url)

        async def _execute():
            if method == 'GET':
                return await asyncio.gather(*[_get(url) for url in urls], return_exceptions=return_exceptions)
            elif method == 'POST':
                return await asyncio.gather(*[_post(url) for url in urls], return_exceptions=return_exceptions)
            else:
                raise ValueError('Invalid method')

        return asyncio.run(_execute())



from dataclasses import dataclass

import httpx


@dataclass
class QMCUser:
    id: int
    name: str


@dataclass
class QMCPage:
    title: str
    unix_name: str


class QuickModule:
    @staticmethod
    def _request(
            module_name: str,
            site_id: int,
            query: str,
    ):
        """リクエストを送信する

        Parameters
        ----------
        module_name: str
            モジュール名
        site_id: int
            サイトID
        query: str
            クエリ
        """

        if module_name not in [
            "MemberLookupQModule",
            "UserLookupQModule",
            "PageLookupQModule",
        ]:
            raise ValueError("Invalid module name")

        url = f'https://www.wikidot.com/quickmodule.php?module={module_name}&s={site_id}&q={query}'
        response = httpx.get(url, timeout=300)
        if response.status_code == httpx.codes.INTERNAL_SERVER_ERROR:
            raise ValueError("Site is not found")
        return response.json()

    @staticmethod
    def member_lookup(
            site_id: int,
            query: str
    ):
        """メンバーを検索する

        Parameters
        ----------
        site_id: int
            サイトID
        query: str
            クエリ

        Returns
        -------
        list[QMCUser]
            ユーザーのリスト
        """
        users = QuickModule._request("MemberLookupQModule", site_id, query)["users"]
        return [QMCUser(id=user["user_id"], name=user["name"]) for user in users]

    @staticmethod
    def user_lookup(
            site_id: int,
            query: str
    ):
        """ユーザーを検索する

        Parameters
        ----------
        site_id: int
            サイトID
        query: str
            クエリ

        Returns
        -------
        list[QMCUser]
            ユーザーのリスト
        """
        users = QuickModule._request("UserLookupQModule", site_id, query)["users"]
        return [QMCUser(id=user["user_id"], name=user["name"]) for user in users]

    @staticmethod
    def page_lookup(
            site_id: int,
            query: str
    ):
        """ページを検索する

        Parameters
        ----------
        site_id: int
            サイトID
        query: str
            クエリ

        Returns
        -------
        list[QMCPage]
            ページのリスト
        """
        pages = QuickModule._request("PageLookupQModule", site_id, query)["pages"]
        return [QMCPage(title=page["title"], unix_name=page["unix_name"]) for page in pages]



from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING

import httpx
from bs4 import BeautifulSoup, Tag, ResultSet

from wikidot.common import exceptions
from wikidot.common.decorators import login_required
from wikidot.util.parser import odate as odate_parser, user as user_parser

if TYPE_CHECKING:
    from wikidot.module.user import AbstractUser, User
    from wikidot.module.client import Client


class PrivateMessageCollection(list):
    def __str__(self):
        return f'{self.__class__.__name__}({len(self)} messages)'

    @staticmethod
    @login_required
    def from_ids(
            client: 'Client',
            message_ids: list[int]
    ) -> 'PrivateMessageCollection':
        """メッセージIDのリストからメッセージオブジェクトのリストを取得する

        Parameters
        ----------
        client: Client
            クライアント
        message_ids: list[int]
            メッセージIDのリスト

        Returns
        -------
        PrivateMessageCollection
            メッセージオブジェクトのリスト
        """
        bodies = []

        for message_id in message_ids:
            bodies.append({
                'item': message_id,
                'moduleName': 'dashboard/messages/DMViewMessageModule'
            })

        responses: [httpx.Response | Exception] = client.amc_client.request(bodies, return_exceptions=True)

        messages = []

        for index, response in enumerate(responses):
            if isinstance(response, exceptions.WikidotStatusCodeException):
                if response.status_code == "no_message":
                    raise exceptions.ForbiddenException(f'Failed to get message: {message_ids[index]}') from response

            if isinstance(response, Exception):
                raise response

            html = BeautifulSoup(response.json()['body'], 'lxml')

            sender, recipient = html.select('div.pmessage div.header span.printuser')

            messages.append(
                PrivateMessage(
                    client=client,
                    id=message_ids[index],
                    sender=user_parser(client, sender),
                    recipient=user_parser(client, recipient),
                    subject=html.select_one('div.pmessage div.header span.subject').get_text(),
                    body=html.select_one('div.pmessage div.body').get_text(),
                    created_at=odate_parser(html.select_one('div.header span.odate'))
                )
            )

        return PrivateMessageCollection(messages)

    @staticmethod
    @login_required
    def _acquire(client: 'Client', module_name: str):
        """受信・送信箱のメッセージを取得する

        Parameters
        ----------
        client: Client
            クライアント
        module_name: str
            モジュール名

        Returns
        -------
        PrivateMessageCollection
            受信箱のメッセージ
        """
        # pager取得
        response = client.amc_client.request([{
            'moduleName': module_name
        }])[0]

        html = BeautifulSoup(response.json()['body'], 'lxml')
        # pagerの最後から2番目の要素を取得
        # pageが存在しない場合は1ページのみ
        pager: ResultSet[Tag] = html.select('div.pager span.target')
        max_page: int = int(pager[-2].get_text()) if len(pager) > 2 else 1

        if max_page > 1:
            # メッセージ取得
            bodies = [{
                'page': page,
                'moduleName': module_name
            } for page in range(1, max_page + 1)]

            responses: [httpx.Response | Exception] = client.amc_client.request(bodies, return_exceptions=False)
        else:
            responses = [response]

        message_ids = []
        for response in responses:
            html = BeautifulSoup(response.json()['body'], 'lxml')
            # tr.messageのdata-href末尾の数字を取得
            message_ids.extend([int(tr['data-href'].split('/')[-1]) for tr in html.select('tr.message')])

        return PrivateMessageCollection.from_ids(client, message_ids)


class PrivateMessageInbox(PrivateMessageCollection):
    @staticmethod
    def from_ids(
            client: 'Client',
            message_ids: list[int]
    ) -> 'PrivateMessageInbox':
        """メッセージIDのリストから受信箱のメッセージオブジェクトのリストを取得する

        Parameters
        ----------
        client: Client
            クライアント
        message_ids: list[int]
            メッセージIDのリスト

        Returns
        -------
        PrivateMessageInbox
            受信箱のメッセージオブジェクトのリスト
        """
        return PrivateMessageInbox(PrivateMessageCollection.from_ids(client, message_ids))

    @staticmethod
    def acquire(client: 'Client'):
        """受信箱のメッセージを取得する

        Parameters
        ----------
        client: Client
            クライアント

        Returns
        -------
        PrivateMessageInbox
            受信箱のメッセージ
        """
        return PrivateMessageInbox(
            PrivateMessageCollection._acquire(client, 'dashboard/messages/DMInboxModule'))


class PrivateMessageSentBox(PrivateMessageCollection):
    @staticmethod
    def from_ids(
            client: 'Client',
            message_ids: list[int]
    ) -> 'PrivateMessageSentBox':
        """メッセージIDのリストから受信箱のメッセージオブジェクトのリストを取得する

        Parameters
        ----------
        client: Client
            クライアント
        message_ids: list[int]
            メッセージIDのリスト

        Returns
        -------
        PrivateMessageSentBox
            受信箱のメッセージオブジェクトのリスト
        """
        return PrivateMessageSentBox(PrivateMessageCollection.from_ids(client, message_ids))

    @staticmethod
    def acquire(client: 'Client') -> 'PrivateMessageSentBox':
        """送信箱のメッセージを取得する

        Parameters
        ----------
        client: Client
            クライアント

        Returns
        -------
        PrivateMessageSentBox
            受信箱のメッセージ
        """
        return PrivateMessageSentBox(PrivateMessageCollection._acquire(client, 'dashboard/messages/DMSentModule'))


@dataclass
class PrivateMessage:
    """プライベートメッセージオブジェクト

    Attributes
    ----------
    client: Client
        クライアントクラスのインスタンス
    id: int
        メッセージID
    sender: AbstractUser
        送信者
    recipient: AbstractUser
        受信者
    subject: str
        件名
    body: str
        本文
    created_at: str
        作成日時
    """
    client: 'Client'
    id: int
    sender: 'AbstractUser'
    recipient: 'AbstractUser'
    subject: str
    body: str
    created_at: datetime

    def __str__(self):
        return f'PrivateMessage(id={self.id}, sender={self.sender}, recipient={self.recipient}, subject={self.subject})'

    @staticmethod
    def from_id(
            client: 'Client',
            message_id: int
    ) -> 'PrivateMessage':
        """メッセージIDからメッセージオブジェクトを取得する

        Parameters
        ----------
        client: Client
            クライアント
        message_id: int
            メッセージID

        Returns
        -------
        PrivateMessage
            メッセージオブジェクト
        """
        return PrivateMessageCollection.from_ids(client, [message_id])[0]

    @staticmethod
    @login_required
    def send(
            client: 'Client',
            recipient: 'User',
            subject: str,
            body: str
    ) -> None:
        """メッセージを送信する

        Parameters
        ----------
        client: Client
            クライアント
        recipient: User
            受信者
        subject: str
            件名
        body: str
            本文
        """
        client.amc_client.request([{
            'source': body,
            'subject': subject,
            'to_user_id': recipient.id,
            'action': 'DashboardMessageAction',
            'event': 'send',
            'moduleName': 'Empty'
        }])



from wikidot.common import wd_logger
from wikidot.connector.ajax import AjaxModuleConnectorClient, AjaxModuleConnectorConfig
from wikidot.module.auth import HTTPAuthentication
from wikidot.module.private_message import PrivateMessage, PrivateMessageInbox, \
    PrivateMessageSentBox, PrivateMessageCollection
from wikidot.module.site import Site
from wikidot.module.user import User, UserCollection


class ClientUserMethods:
    def __init__(self, client: 'Client'):
        self.client = client

    def get(
            self,
            name: str,
            raise_when_not_found: bool = False
    ) -> User:
        """ユーザー名からユーザーオブジェクトを取得する

        Parameters
        ----------
        name: str
            ユーザー名

        raise_when_not_found: bool
            ユーザーが見つからない場合に例外を送出するかどうか (True: 送出する, False: 送出しない)
            デフォルトでは送出せずにNoneを返す

        Returns
        -------
        User
            ユーザーオブジェクト
        """
        return User.from_name(self.client, name, raise_when_not_found)

    def get_bulk(
            self,
            names: list[str],
            raise_when_not_found: bool = False
    ) -> list[User]:
        """ユーザー名からユーザーオブジェクトを取得する

        Parameters
        ----------
        names: list[str]
            ユーザー名のリスト
        raise_when_not_found: bool
            ユーザーが見つからない場合に例外を送出するかどうか (True: 送出する, False: 送出しない)
            デフォルトでは送出せずにNoneを返す

        Returns
        -------
        list[User]
            ユーザーオブジェクトのリスト
        """
        return UserCollection.from_names(self.client, names, raise_when_not_found)


class ClientPrivateMessageMethods:
    def __init__(self, client: 'Client'):
        self.client = client

    def send(
            self,
            recipient: User,
            subject: str,
            body: str
    ) -> None:
        """メッセージを送信する

        Parameters
        ----------
        recipient: User
            受信者
        subject: str
            件名
        body: str
            本文
        """
        PrivateMessage.send(self.client, recipient, subject, body)

    def get_inbox(self) -> PrivateMessageInbox:
        """受信箱を取得する

        Returns
        -------
        PrivateMessageInbox
            受信箱
        """
        return PrivateMessageInbox.acquire(self.client)

    def get_sentbox(self) -> PrivateMessageSentBox:
        """送信箱を取得する

        Returns
        -------
        PrivateMessageSentBox
            送信箱
        """
        return PrivateMessageSentBox.acquire(self.client)

    def get_messages(self, message_ids: list[int]) -> PrivateMessageCollection:
        """メッセージを取得する

        Parameters
        ----------
        message_ids: list[int]
            メッセージIDのリスト

        Returns
        -------
        list[PrivateMessage]
            メッセージのリスト
        """
        return PrivateMessageCollection.from_ids(self.client, message_ids)

    def get_message(self, message_id: int) -> PrivateMessage:
        """メッセージを取得する

        Parameters
        ----------
        message_id: int
            メッセージID

        Returns
        -------
        PrivateMessage
            メッセージ
        """
        return PrivateMessage.from_id(self.client, message_id)


class ClientSiteMethods:
    def __init__(self, client: 'Client'):
        self.client = client

    def get(self, unix_name: str) -> 'Site':
        """UNIX名からサイトオブジェクトを取得する

        Parameters
        ----------
        unix_name: str
            サイトのUNIX名

        Returns
        -------
        Site
            サイトオブジェクト
        """
        return Site.from_unix_name(self.client, unix_name)


class Client:
    """基幹クライアント"""

    def __init__(
            self,
            username: str | None = None,
            password: str | None = None,
            amc_config: AjaxModuleConnectorConfig | None = None,
            logging_level: str = "WARNING"
    ):
        """Wikidot Client

        Parameters
        ----------
        username: str | None
            ユーザー名
        password: str | None
            パスワード
        amc_config: dict | None
            AMCの設定
        logging_level: str
            ロギングレベル
        """
        # 最初にロギングレベルを決定する
        wd_logger.setLevel(logging_level)

        # AMCClientを初期化
        self.amc_client = AjaxModuleConnectorClient(
            site_name=None,
            config=amc_config
        )

        # セッション関連変数の初期化
        self.is_logged_in = False
        self.username = None

        # usernameとpasswordが指定されていればログインする
        if username is not None and password is not None:
            HTTPAuthentication.login(self, username, password)
            self.is_logged_in = True
            self.username = username

        # ----------
        # 以下メソッド
        # ----------

        self.user = ClientUserMethods(self)
        self.private_message = ClientPrivateMessageMethods(self)
        self.site = ClientSiteMethods(self)

        # ------------
        # メソッド終わり
        # ------------

    def __del__(self):
        """デストラクタ"""
        if self.is_logged_in:
            HTTPAuthentication.logout(self)
            self.is_logged_in = False
            self.username = None
        del self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__del__()
        return

    def __str__(self):
        return f"Client(username={self.username}, is_logged_in={self.is_logged_in})"



import asyncio
import json.decoder
from dataclasses import dataclass

import httpx

from ..common import wd_logger
from ..common.exceptions import NotFoundException, AMCHttpStatusCodeException, ResponseDataException, \
    WikidotStatusCodeException


class AjaxRequestHeader:
    """ajax-module-connector.phpへのリクエスト時に利用するヘッダの構築用クラス"""

    def __init__(self,
                 content_type: str | None = None,
                 user_agent: str | None = None,
                 referer: str | None = None,
                 cookie: dict | None = None):
        """AjaxRequestHeaderオブジェクトの初期化

        Parameters
        ----------
        content_type: str | None
            Content-Type
        user_agent: str | None
            User-Agent
        referer: str | None
            Referer
        cookie: dict | None
            Cookie
        """
        self.content_type: str = "application/x-www-form-urlencoded; charset=UTF-8" if content_type is None else content_type
        self.user_agent: str = "WikidotPy" if user_agent is None else user_agent
        self.referer: str = "https://www.wikidot.com/" if referer is None else referer
        self.cookie: dict[str, any] = {
            "wikidot_token7": 123456
        }
        if cookie is not None:
            self.cookie.update(cookie)
        return

    def set_cookie(self, name, value) -> None:
        """Cookieを設定する

        Parameters
        ----------
        name: str
            Cookie名
        value: str
            Cookie値
        """
        self.cookie[name] = value
        return

    def delete_cookie(self, name) -> None:
        """Cookieを削除する

        Parameters
        ----------
        name: str
            Cookie名
        """
        del self.cookie[name]
        return

    def is_cookie_set(self, name) -> bool:
        """Cookieが設定されているかどうかを返す

        Parameters
        ----------
        name: str
            Cookie名

        Returns
        -------
        bool
            Cookieが設定されているかどうか
        """
        return name in self.cookie

    def get_header(self) -> dict:
        """ヘッダを構築して返す

        Returns
        -------
        dict
            ヘッダ
        """
        return {
            "Content-Type": self.content_type,
            "User-Agent": self.user_agent,
            "Referer": self.referer,
            "Cookie": "".join([f'{name}={value};' for name, value in self.cookie.items()])
        }


@dataclass
class AjaxModuleConnectorConfig:
    """ajax-module-connector.phpへのリクエストを行う際の設定

    Attributes
    ----------
    request_timeout: int
        タイムアウト
    attempt_limit: int
        リクエストの試行回数上限
    retry_interval: int
        リクエスト失敗時のリトライ間隔
    semaphore_limit: int
        セマフォの上限
    """
    request_timeout: int = 20
    attempt_limit: int = 3
    retry_interval: int = 5
    semaphore_limit: int = 10


class AjaxModuleConnectorClient:
    """ajax-module-connector.phpへのリクエストを行うクライアント"""

    def __init__(
            self,
            site_name: str | None = None,
            config: AjaxModuleConnectorConfig | None = None
    ):
        """AjaxModuleConnectorClientオブジェクトの初期化

        Parameters
        ----------
        config: AjaxModuleConnectorConfig
            設定
        """
        self.site_name: str = site_name if site_name is not None else 'www'
        self.config: AjaxModuleConnectorConfig = config if config is not None else AjaxModuleConnectorConfig()

        # ssl対応チェック
        self.ssl_supported: bool = self._check_existence_and_ssl()

        # ヘッダの初期化
        self.header: AjaxRequestHeader = AjaxRequestHeader()

    def _check_existence_and_ssl(self):
        """実際にアクセスしてみてサイトの存在とSSL対応をチェック"""

        # wwwは常にSSL対応
        if self.site_name == 'www':
            return True

        # それ以外のサイトはhttpsにリダイレクトされるかどうかで判断
        response = httpx.get(f'http://{self.site_name}.wikidot.com')

        # 存在しなければ例外送出
        if response.status_code == httpx.codes.NOT_FOUND:
            raise NotFoundException(f'Site is not found: {self.site_name}.wikidot.com')

        # httpsにリダイレクトされているかどうかで判断
        return response.status_code == httpx.codes.MOVED_PERMANENTLY and 'Location' in response.headers and \
            response.headers['Location'].startswith('https')

    def request(
            self,
            bodies: list[dict[str, any]],
            return_exceptions: bool = False,
            site_name: str | None = None,
            site_ssl_supported: bool | None = None
    ) -> list[httpx.Response | Exception]:
        """ajax-module-connector.phpへのリクエストを行う

        Parameters
        ----------
        bodies: list[dict]
            リクエストボディのリスト
        return_exceptions: bool
            例外を返すかどうか (True: 返す, False: 例外を送出)
            デフォルトでは例外を送出
        site_name: str | None
            サイト名
            デフォルトでは初期化時に指定したサイト名
        site_ssl_supported: bool | None
            サイトがSSL対応しているかどうか
            デフォルトでは初期化時にチェックした結果

        Returns
        -------
        dict
            レスポンスボディのリスト（順序はリクエストボディのリストと同じ）

        Raises
        ------
        AMCHttpStatusCodeException
            AMCから返却されたHTTPステータスが200以外だったときの例外

        WikidotStatusCodeException
            AMCから返却されたデータ内のステータスがokではなかったときの例外
            HTTPステータスが200以外の場合はAMCHttpStatusCodeExceptionを投げる

        ResponseDataException
            AMCから返却されたデータが不正だったときの例外
            JSONデータとしてパースできなかった場合や空だった場合に投げる
        """
        semaphore_instance = asyncio.Semaphore(self.config.semaphore_limit)

        site_name = site_name if site_name is not None else self.site_name
        site_ssl_supported = site_ssl_supported if site_ssl_supported is not None else self.ssl_supported

        async def _request(
                _body: dict[str, any]
        ) -> httpx.Response:
            retry_count = 0

            while True:

                # リクエスト実行
                try:
                    # Semaphoreで同時実行数制御
                    async with semaphore_instance:
                        async with httpx.AsyncClient() as client:
                            url = (
                                f'http{"s" if site_ssl_supported else ""}://{site_name}.wikidot.com/'
                                f'ajax-module-connector.php')
                            _body['wikidot_token7'] = 123456
                            wd_logger.debug(f'Ajax Request: {url} -> {_body}')
                            response = await client.post(
                                url,
                                headers=self.header.get_header(),
                                data=_body,
                                timeout=self.config.request_timeout
                            )
                            response.raise_for_status()
                except (httpx.HTTPStatusError, httpx.TimeoutException) as e:
                    # HTTPステータスエラーまたはタイムアウトの場合はリトライ
                    retry_count += 1

                    # リトライ回数上限に達した場合は例外送出
                    if retry_count >= self.config.attempt_limit:
                        wd_logger.error(f'AMC is respond HTTP error code: {response.status_code} -> {_body}')
                        raise AMCHttpStatusCodeException(
                            f'AMC is respond HTTP error code: {response.status_code}',
                            response.status_code
                        ) from e

                    # 間隔を空けてリトライ
                    wd_logger.info(f'AMC is respond status: {response.status_code} (retry: {retry_count}) -> {_body}')
                    await asyncio.sleep(self.config.retry_interval)
                    continue

                # bodyをJSONデータとしてパース
                try:
                    _response_body = response.json()
                except json.decoder.JSONDecodeError as e:
                    # パースできなかったらエラーとして扱う
                    wd_logger.error(f'AMC is respond non-json data: "{response.text}" -> {_body}')
                    raise ResponseDataException(f'AMC is respond non-json data: "{response.text}"') from e

                # レスポンスが空だったらエラーとして扱う
                if _response_body is None or len(_response_body) == 0:
                    wd_logger.error(f'AMC is respond empty data -> {_body}')
                    raise ResponseDataException(f'AMC is respond empty data')

                # 中身のstatusがokでなかったらエラーとして扱う
                if 'status' in _response_body:
                    # statusがtry_againの場合はリトライ
                    if _response_body['status'] == 'try_again':
                        retry_count += 1
                        if retry_count >= self.config.attempt_limit:
                            wd_logger.error(f'AMC is respond status: "try_again" -> {_body}')
                            raise WikidotStatusCodeException(f'AMC is respond status: "try_again"', 'try_again')

                        wd_logger.info(f'AMC is respond status: "try_again" (retry: {retry_count})')
                        await asyncio.sleep(self.config.retry_interval)
                        continue

                    # それ以外でstatusがokでない場合はエラーとして扱う
                    elif _response_body['status'] != 'ok':
                        wd_logger.error(f'AMC is respond error status: "{_response_body["status"]}" -> {_body}')
                        raise WikidotStatusCodeException(
                            f'AMC is respond error status: "{_response_body["status"]}"',
                            _response_body['status']
                        )

                # レスポンスを返す
                return response

        async def _execute_requests():
            return await asyncio.gather(*[_request(body) for body in bodies], return_exceptions=return_exceptions)

        # 処理を実行
        return asyncio.run(_execute_requests())
